/*
 * Copyright (C) 2015 Darron Broad
 * All rights reserved.
 * 
 * This file is part of Ape65.
 * 
 * Ape65 is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation. 
 * 
 * Ape65 is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details. 
 * 
 * You should have received a copy of the GNU General Public License along
 * with Ape65. If not, see http://www.gnu.org/licenses/
 */

#include "ape65.h"

/*
 * Instruction [0..55] to Mnemonic
 */
static const char *_mnemonic[NINSTR] = {
	"ADC", "AND", "ASL", "BCC", "BCS", "BEQ",
	"BIT", "BMI", "BNE", "BPL", "BRK", "BVC",
	"BVS", "CLC", "CLD", "CLI", "CLV", "CMP",
	"CPX", "CPY", "DEC", "DEX", "DEY", "EOR",
	"INC", "INX", "INY", "JMP", "JSR", "LDA",
	"LDX", "LDY", "LSR", "NOP", "ORA", "PHA",
	"PHP", "PLA", "PLP", "ROL", "ROR", "RTI",
	"RTS", "SBC", "SEC", "SED", "SEI", "STA",
	"STX", "STY", "TAX", "TAY", "TSX", "TXA",
	"TXS", "TYA"
};

/*
 * Instruction [0..55] + Addressing mode [0..12] to Opcode [0..255]
 */
static const uint8_t _opcode[NINSTR * NMODES] = {
	FF, FF, 0x69, 0x65, 0x75, FF, 0x6d, 0x7d, 0x79, 0x61, 0x71, FF, FF,	//  0 ADC
	FF, FF, 0x29, 0x25, 0x35, FF, 0x2d, 0x3d, 0x39, 0x21, 0x31, FF, FF,	//  1 AND
	FF, 0x0a, FF, 0x06, 0x16, FF, 0x0e, 0x1e, FF, FF, FF, FF, FF,		//  2 ASL
	FF, FF, FF, FF, FF, FF, FF, FF, FF, FF, FF, 0x90, FF,			//  3 BCC
	FF, FF, FF, FF, FF, FF, FF, FF, FF, FF, FF, 0xb0, FF,			//  4 BCS
	FF, FF, FF, FF, FF, FF, FF, FF, FF, FF, FF, 0xf0, FF,			//  5 BEQ
	FF, FF, FF, 0x24, FF, FF, 0x2c, FF, FF, FF, FF, FF, FF,			//  6 BIT
	FF, FF, FF, FF, FF, FF, FF, FF, FF, FF, FF, 0x30, FF,			//  7 BMI
	FF, FF, FF, FF, FF, FF, FF, FF, FF, FF, FF, 0xd0, FF,			//  8 BNE
	FF, FF, FF, FF, FF, FF, FF, FF, FF, FF, FF, 0x10, FF,			//  9 BPL
	0x00, FF, FF, FF, FF, FF, FF, FF, FF, FF, FF, FF, FF,			// 10 BRK
	FF, FF, FF, FF, FF, FF, FF, FF, FF, FF, FF, 0x50, FF,			// 11 BVC
	FF, FF, FF, FF, FF, FF, FF, FF, FF, FF, FF, 0x70, FF,			// 12 BVS
	0x18, FF, FF, FF, FF, FF, FF, FF, FF, FF, FF, FF, FF,			// 13 CLC
	0xd8, FF, FF, FF, FF, FF, FF, FF, FF, FF, FF, FF, FF,			// 14 CLD
	0x58, FF, FF, FF, FF, FF, FF, FF, FF, FF, FF, FF, FF,			// 15 CLI
	0xb8, FF, FF, FF, FF, FF, FF, FF, FF, FF, FF, FF, FF,			// 16 CLV
	FF, FF, 0xc9, 0xc5, 0xd5, FF, 0xcd, 0xdd, 0xd9, 0xc1, 0xd1, FF, FF,	// 17 CMP
	FF, FF, 0xe0, 0xe4, FF, FF, 0xec, FF, FF, FF, FF, FF, FF,		// 18 CPX
	FF, FF, 0xc0, 0xc4, FF, FF, 0xcc, FF, FF, FF, FF, FF, FF,		// 19 CPY
	FF, FF, FF, 0xc6, 0xd6, FF, 0xce, 0xde, FF, FF, FF, FF, FF,		// 20 DEC
	0xca, FF, FF, FF, FF, FF, FF, FF, FF, FF, FF, FF, FF,			// 21 DEX
	0x88, FF, FF, FF, FF, FF, FF, FF, FF, FF, FF, FF, FF,			// 22 DEY
	FF, FF, 0x49, 0x45, 0x55, FF, 0x4d, 0x5d, 0x59, 0x41, 0x51, FF, FF,	// 23 EOR
	FF, FF, FF, 0xe6, 0xf6, FF, 0xee, 0xfe, FF, FF, FF, FF, FF,		// 24 INC
	0xe8, FF, FF, FF, FF, FF, FF, FF, FF, FF, FF, FF, FF,			// 25 INX
	0xc8, FF, FF, FF, FF, FF, FF, FF, FF, FF, FF, FF, FF,			// 26 INY
	FF, FF, FF, FF, FF, FF, 0x4c, FF, FF, FF, FF, FF, 0x6c,			// 27 JMP
	FF, FF, FF, FF, FF, FF, 0X20, FF, FF, FF, FF, FF, FF,			// 28 JSR
	FF, FF, 0xa9, 0xa5, 0xb5, FF, 0xad, 0xbd, 0xb9, 0xa1, 0xb1, FF, FF,	// 29 LDA
	FF, FF, 0xa2, 0xa6, FF, 0xb6, 0xae, FF, 0xbe, FF, FF, FF, FF,		// 30 LDX
	FF, FF, 0xa0, 0xa4, 0xb4, FF, 0xac, 0xbc, FF, FF, FF, FF, FF,		// 31 LDY
	FF, 0x4a, FF, 0x46, 0x56, FF, 0x4e, 0x5e, FF, FF, FF, FF, FF,		// 32 LSR
	0xea, FF, FF, FF, FF, FF, FF, FF, FF, FF, FF, FF, FF,			// 33 NOP
	FF, FF, 0x09, 0x05, 0x15, FF, 0x0d, 0x1d, 0x19, 0x01, 0x11, FF, FF,	// 34 ORA
	0x48, FF, FF, FF, FF, FF, FF, FF, FF, FF, FF, FF, FF,			// 35 PHA
	0x08, FF, FF, FF, FF, FF, FF, FF, FF, FF, FF, FF, FF,			// 36 PHP
	0x68, FF, FF, FF, FF, FF, FF, FF, FF, FF, FF, FF, FF,			// 37 PLA
	0x28, FF, FF, FF, FF, FF, FF, FF, FF, FF, FF, FF, FF,			// 38 PLP
	FF, 0x2a, FF, 0x26, 0x36, FF, 0x2e, 0x3e, FF, FF, FF, FF, FF,		// 39 ROL
	FF, 0x6a, FF, 0x66, 0x76, FF, 0x6e, 0x7e, FF, FF, FF, FF, FF,		// 40 ROR
	0x40, FF, FF, FF, FF, FF, FF, FF, FF, FF, FF, FF, FF,			// 41 RTI
	0x60, FF, FF, FF, FF, FF, FF, FF, FF, FF, FF, FF, FF,			// 42 RTS
	FF, FF, 0xe9, 0xe5, 0xf5, FF, 0xed, 0xfd, 0xf9, 0xe1, 0xf1, FF, FF,	// 43 SBC
	0x38, FF, FF, FF, FF, FF, FF, FF, FF, FF, FF, FF, FF,			// 44 SEC
	0xf8, FF, FF, FF, FF, FF, FF, FF, FF, FF, FF, FF, FF,			// 45 SED
	0x78, FF, FF, FF, FF, FF, FF, FF, FF, FF, FF, FF, FF,			// 46 SEI
	FF, FF, FF, 0x85, 0x95, FF, 0x8d, 0x9d, 0x99, 0x81, 0x91, FF, FF,	// 47 STA
	FF, FF, FF, 0x86, FF, 0x96, 0x8e, FF, FF, FF, FF, FF, FF,		// 48 STX
	FF, FF, FF, 0x84, 0x94, FF, 0x8c, FF, FF, FF, FF, FF, FF,		// 49 STY
	0xaa, FF, FF, FF, FF, FF, FF, FF, FF, FF, FF, FF, FF,			// 50 TAX
	0xa8, FF, FF, FF, FF, FF, FF, FF, FF, FF, FF, FF, FF,			// 51 TAY
	0xba, FF, FF, FF, FF, FF, FF, FF, FF, FF, FF, FF, FF,			// 52 TSX
	0x8a, FF, FF, FF, FF, FF, FF, FF, FF, FF, FF, FF, FF,			// 53 TXA
	0x9a, FF, FF, FF, FF, FF, FF, FF, FF, FF, FF, FF, FF,			// 54 TXS
	0x98, FF, FF, FF, FF, FF, FF, FF, FF, FF, FF, FF, FF			// 55 TYA
};

/*
 * Opcode [0..255] to Instruction [0..55]
 */
extern const uint8_t _instruction[256];

/*
 * Opcode [0..255] to Addressing mode [0..12]
 */
extern const uint8_t _mode[256];

/*
 * Addressing mode [0..12] to Number of bytes [1..3]
 */
static const uint8_t _bytes[NMODES] = {1, 1, 2, 2, 2, 2, 3, 3, 3, 2, 2, 2, 3}; 

/*
 * Returns
 * 	0..55		Mnemonic index
 * 	0xFF		Not found
 */
static inline uint8_t
mnemonic(char *str)
{
	int8_t i;

	for (i = 0; i < NINSTR; ++i) {
		if (mystrcasestr(str, _mnemonic[i]) == str)
			return (i);
	}

	return (FF);
}

/*
 * Returns
 * 	Operand
 *
 * Eg.
 * 	180   => 180
 * 	$FF   => 255
 * 	%1111 => 15
 *
 * % = binary
 * @ = octal
 * $ = hexadecimal
 */
static uint16_t
operand(char **str)
{
	uint8_t base = 10;

	skipspace(str);

	if (**str == '%') {
		base = 2;
		(*str)++;
	} else if (**str == '@') {
		base = 8;
		(*str)++;
	} else if (**str == '$') {
		base = 16;
		(*str)++;
	}

	return (strtoint(str, base));
}

/*
 * Returns
 * 	0..12		Addressing mode of instruction
 * 	0xFF		Syntax error
 *
 * Warning
 * 	This function will return ZEROPAGE when the operand
 * 	is 8-bits. This mode may be invalid for the operation
 * 	and the mode may really be ABSOLUTE.
 */
static inline uint8_t
getmode(uint8_t mne, char *str, uint16_t addr, uint16_t *data)
{
	if (*str == '\0')
		return (IMPLIED);

	if (*str == '#') {
		str++;
		*data = operand(&str);
		return (IMMEDIATE);
	}

	if (*str == '(') {
		str++;
		*data = operand(&str);
		if (*str == ',' && (str[1] == 'x' || str[1] == 'X') && str[2] == ')')
			return (INDIRECTX);
		if (*str == ')' && str[1] == ',' && (str[2] == 'y' || str[2] == 'Y'))
			return (INDIRECTY);
		return (ABSOLUTEI);
	}

	if (*str == 'A' && str[1] == '\0') {
		return (ACCUMULATOR);
	}

	*data = operand(&str);

	/* Relative branch */
	if (mne >= 3 && mne <= 12 && mne != 6 && mne != 10) {
		addr += 2;
		if (*data >= addr) {	/* Forward */
			if ((*data - addr) > 127)
				return (FF);
		} else {		/* Backward */
			if ((addr - *data) > 128)
				return (FF);
		}
		*data = (*data - addr) & 0xFF;
		return (RELATIVE);
	}

	if (*str == '\0') {
		if (*data > 0xFF)
			return (ABSOLUTE);
		else
			return (ZEROPAGE);
	}

	if (*str == ',') {
		if (str[1] == 'x' || str[1] == 'X') {
			if (*data > 0xFF)
				return (ABSOLUTEX);
			else
				return (ZEROPAGEX);
		}
		if (str[1] == 'y' || str[1] == 'Y') {
			if (*data > 0xFF)
				return (ABSOLUTEY);
			else
				return (ZEROPAGEY);
		}
	}

	return (FF);
}

/*
 * Returns
 * 	0xFF		Error
 *
 * Side effects
 * 	Will promote mode from ZEROPAGE to ABSOLUTE if
 * 	ZEROPAGE fails.
 */
static inline uint8_t
getopcode(uint8_t mne, uint8_t *mode)
{
	uint8_t op = _opcode[mne * NMODES + *mode];

	if (op == 0xFF) {
		/* Promote zero page to absolute */
		if (*mode == ZEROPAGEX)
			*mode = ABSOLUTEX;
		else if (*mode == ZEROPAGEY)
			*mode = ABSOLUTEY;
		else if (*mode == ZEROPAGE)
			*mode = ABSOLUTE;

		/* Try again */
		op = _opcode[mne * NMODES + *mode];
	}

	return (op);
}

/*
 * Returns
 * 	0xFF		Error
 */
static inline uint8_t
getbytes(uint8_t mode)
{
	if (mode >= NMODES)
		return 0xFF;

	return (_bytes[mode]);
}

/*
 * Returns
 * 	false		Okay
 * 	true		Syntax error
 *
 * 	bytes[0]	opcode
 * 	bytes[1]	address or data
 * 	bytes[2]	address high byte
 */
uint8_t
assembler(uint16_t addr, char *str, uint8_t bytes[3], uint8_t *nbytes)
{
	uint8_t i, opcode, mode;
	uint16_t data = 0;

	i = mnemonic(str);
	if (i == 0xFF)
		return true;

	str += 3;
	skipspace(&str);

	mode = getmode(i, str, addr, &data);
	if (mode == 0xFF)
		return true;

	opcode = getopcode(i, &mode);
	if (opcode == 0xFF)
		return true;

	*nbytes = getbytes(mode);

	switch (*nbytes) {
	case 3: bytes[2] = data >> 8;
	case 2: bytes[1] = data;
	case 1: bytes[0] = opcode;
		break;
	default:return true;
	}

	return false;
}

/*
 * Returns
 * 	false		Ok
 * 	true		Error
 */
uint8_t
disassembler(uint16_t addr, char *str, uint8_t bytes[3], uint8_t *nbytes)
{
	uint8_t opcode, mode, instr;
	const char *mne;

	opcode = bytes[0];		/* 0..255 */
	instr = _instruction[opcode];	/* 0..55 */
	mode = _mode[opcode];		/* 0..12 */

	if (instr == 0xFF || mode == 0xFF) {
		snprintf(str, SLEN, "%02X       ???",
			opcode);
		*nbytes = 1;
		return (false);
	}

	*nbytes = getbytes(mode);

	mne = _mnemonic[instr];

	switch (mode) {

	case IMPLIED:		/* CLC */
		snprintf(str, SLEN, "%02X       %s",
			opcode, mne);
		break;

	case ACCUMULATOR:	/* ASL A */
		snprintf(str, SLEN, "%02X       %s A",
			opcode, mne);
		break;

	case IMMEDIATE:		/* LDA #$12 */
		snprintf(str, SLEN, "%02X %02X    %s #$%02X",
			opcode, bytes[1], mne, bytes[1]);
		break;

	case ZEROPAGE:		/* LDA $12 */
		snprintf(str, SLEN, "%02X %02X    %s $%02X",
			opcode, bytes[1], mne, bytes[1]);
		break;

	case ZEROPAGEX:		/* LDA $12,X */
		snprintf(str, SLEN, "%02X %02X    %s $%02X,X",
			opcode, bytes[1], mne, bytes[1]);
		break;

	case ZEROPAGEY:		/* LDA $12,Y */
		snprintf(str, SLEN, "%02X %02X    %s $%02X,Y",
			opcode, bytes[1], mne, bytes[1]);
		break;

	case ABSOLUTE:		/* JMP $1234 */
		snprintf(str, SLEN, "%02X %02X %02X %s $%04X",
			opcode, bytes[1], bytes[2], mne, bytes[1] | bytes[2] << 8);
		break;

	case ABSOLUTEX:		/* LDA $1234,X */
		snprintf(str, SLEN, "%02X %02X %02X %s $%04X,X",
			opcode, bytes[1], bytes[2], mne, bytes[1] | bytes[2] << 8);
		break;

	case ABSOLUTEY:		/* LDA $1234,Y */
		snprintf(str, SLEN, "%02X %02X %02X %s $%04X,Y",
			opcode, bytes[1], bytes[2], mne, bytes[1] | bytes[2] << 8);
		break;

	case INDIRECTX:		/* LDA ($12,X) */
		snprintf(str, SLEN, "%02X %02X    %s ($%02X,X)",
			opcode, bytes[1], mne, bytes[1]);
		break;

	case INDIRECTY:		/* LDA ($12),Y */
		snprintf(str, SLEN, "%02X %02X    %s ($%02X),Y",
			opcode, bytes[1], mne, bytes[1]);
		break;

	case RELATIVE:		/* BCC $1234 */
		addr += (int8_t)bytes[1] + 2;
		snprintf(str, SLEN, "%02X %02X    %s $%04X",
			opcode, bytes[1], mne, addr);
		break;

	case ABSOLUTEI:		/* JMP ($1234) */
		snprintf(str, SLEN, "%02X %02X %02X %s ($%04X)",
			opcode, bytes[1], bytes[2], mne, bytes[1] | bytes[2] << 8);
	default:break;
	}

	return (false);
}
