/*
 * Copyright (C) 2015 Darron Broad
 * All rights reserved.
 * 
 * This file is part of Ape65.
 * 
 * Ape65 is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation. 
 * 
 * Ape65 is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details. 
 * 
 * You should have received a copy of the GNU General Public License along
 * with Ape65. If not, see http://www.gnu.org/licenses/
 */

/*
 * This emulates the TEDMON machine code monitor found on the Commodore PLUS/4
 * with a few additional commands and minor enhancements.
 *
 * TEDMON commands are emulated exactly as the original except for save, load
 * and verify which do not support I/O device selection.
 */

#include "ape65.h"

struct monitor_state mon = {0};

extern struct register_file reg;

extern char runstop;

static inline char
getrunstop(void)
{
	return getcharacter();
}

static void
registers(uint8_t opt)
{
	if (opt & BANNER)
		printf("   PC  SR AC XR YR SP NV.BDIZC\r\n");

	if (opt & REGISTERS) {
		uint8_t bit = 0x80;

		printf("; %04X %02X %02X %02X %02X %02X ",
			reg.PC, reg.SR, reg.AC, reg.XR, reg.YR, reg.SP);
		while (bit) {
			if (reg.SR & bit)
				putchar('1');
			else
				putchar('.');
			bit >>= 1;
		}
	}

	if (opt & NEWLINE)
		printf("\r\n");
}

static void
trace(void)
{
	uint8_t banner = -1;

	do {
		char str[SLEN];
		uint8_t bytes[3];
		uint8_t nbytes;

		if (banner++ >= 0x14) {
			registers(BANNER);
			banner = 0;
		}
		registers(REGISTERS);

		bytes[0] = peek(reg.PC);
		bytes[1] = peek(reg.PC + 1);
		bytes[2] = peek(reg.PC + 2);
		disassembler(reg.PC, str, bytes, &nbytes);
		printf(" %s\r\n", str);
	}
	while (mpu() == 0 && getrunstop() != 3);
}

#if 0
static void
irq(void)
{
	if (reg.SR & INTERRUPT) {
		printf("IRQ disabled\r\n");
		return;
	}

	_IRQ();
	trace();
	reg.SR &= ~INTERRUPT;
}

static void
nmi(void)
{
	uint8_t SR = reg.SR & INTERRUPT;

	_NMI();
	trace();
	reg.SR &= ~INTERRUPT;
	reg.SR |= SR;
}
#endif

static void
convert(uint16_t number)
{
	char str[SLEN];

	/* 16 digit binary */
	printf("%%%s ", tostring(str, SLEN, number, 2, 16));

	/* 6 digit octal */
	printf("@%s ", tostring(str, SLEN, number, 8, 6));

	/* 5 digit decimal */
	printf("#%s ", tostring(str, SLEN, number, 10, 5));

	/* 4 digit hexadecimal */
	printf("$%s\r\n", tostring(str, SLEN, number, 16, 4));
}

static uint8_t
memory_dump(uint16_t *f, uint16_t l)
{
	uint8_t found = false, i, m;
	char db[12]; /* Data byte :XXXXXXXX */

	db[0] = ':';
	db[9] = '\r';
	db[10] = '\n';
	db[11] = '\0';

	printf(">%04X ", *f);
	for (i = 1; i < 9; ++i) {
		if (*f == l)
			found = true;
		m = peek((*f)++);
		printf("%02X ", m);
		if (m < ' ' || m > 126)
			db[i] = '.';
		else
			db[i] = m;
	}
	printf("%s", db);

	return found;
}

static void
memory(char *str)
{
	uint16_t f;
	uint16_t l;
	uint8_t found;

	/* First address */
	if (*str)
		f = strtoint(&str, 16);
	else
		f = mon.addr;

	/* Last address */
	if (*str)
		l = strtoint(&str, 16);
	else
		l = f + (11 * 8);

	found = false;
	while (found == false && getrunstop() != 3)
		found = memory_dump(&f, l);

	mon.addr = f;
}

/*
 * Returns
 * 	false		Okay
 * 	true		Syntax error
 */
static uint8_t
fill(char *str)
{
	uint16_t f;
	uint16_t l;
	uint8_t d;

	/* First address */
	if (*str)
		f = strtoint(&str, 16);
	else
		return (true);

	/* Last address */
	if (*str)
		l = strtoint(&str, 16);
	else
		return (true);

	/* Data */
	if (*str)
		d = strtoint(&str, 16);
	else
		return (true);

	while (f != l)
		poke(f++, d);
	poke(f++, d);

	mon.addr = f;

	return (false);
}

static void
modify(char *str)
{
	uint16_t a;

	/* Address */
	if (*str) {
		a = strtoint(&str, 16);
		mon.addr = a;
	}
	else
		a = mon.addr;

	while (*str) {
		uint8_t d = strtoint(&str, 16);
		poke(a++, d);
	}

	printf("\033[1A\033[0K");
	a = mon.addr;
	memory_dump(&a, a);
}

/*
 * Returns
 * 	false		Okay
 * 	true		Syntax error
 */
static uint8_t
transfer(char *str)
{
	uint16_t f;
	uint16_t l;
	uint16_t d;

	/* First address */
	if (*str)
		f = strtoint(&str, 16);
	else
		return (true);

	/* Last address */
	if (*str)
		l = strtoint(&str, 16);
	else
		return (true);

	/* Destination address */
	if (*str)
		d = strtoint(&str, 16);
	else
		return (true);

	while (f != l && getrunstop() != 3)
		poke(d++, peek(f++));
	poke(d++, peek(f++));

	mon.addr = f;

	return (false);
}

/*
 * Returns
 * 	false		Okay
 * 	true		Syntax error
 */
static uint8_t
compare(char *str)
{
	uint16_t f;
	uint16_t l;
	uint16_t d;

	/* First address */
	if (*str)
		f = strtoint(&str, 16);
	else
		return (true);

	/* Last address */
	if (*str)
		l = strtoint(&str, 16);
	else
		return (true);

	/* Destination address */
	if (*str)
		d = strtoint(&str, 16);
	else
		return (true);

	while (f != l && getrunstop() != 3) {
		if (peek(f++) != peek(d++))
			printf("%04X ", f - 1);
	}
	if (peek(f++) != peek(d++))
		printf("%04X ", f - 1);
	printf("\r\n");

	mon.addr = f;

	return (false);
}

/*
 * Returns
 * 	false		Okay
 * 	true		Syntax error
 */
static uint8_t
hunt(char *str)
{
	uint16_t f;
	uint16_t l;

	uint8_t hunt[SLEN];
	uint16_t n = 0, i;

	/* First address */
	if (*str)
		f = strtoint(&str, 16);
	else
		return (true);

	/* Last address */
	if (*str)
		l = strtoint(&str, 16);
	else
		return (true);
	
	/* Data? */
	if (*str == '\0')
		return (true);

	/* Data... */
	if (*str == '"' || *str == '\'') {
		str++;
		while (*str)
			hunt[n++] = *str++;
	} else {
		while (*str)
			hunt[n++] = strtoint(&str, 16);
	}

	/* Release the hounds... */
	while (f != l && getrunstop() != 3) {
		for (i = 0; i < n; ++i) {
			if (peek(f + i) != hunt[i])
				break;
		}
		if (i == n)
			printf("%04X ", f);
		f++;
	}
	for (i = 0; i < n; ++i) {
		if (peek(f + i) != hunt[i])
			break;
	}
	if (i == n)
		printf("%04X ", f);
	f++;
	printf("\r\n");

	mon.addr = f;

	return (false);
}

static void
setreg(char *str)
{
	if (*str)
		reg.PC = strtoint(&str, 16);
	if (*str)
		reg.SR = strtoint(&str, 16);
	if (*str)
		reg.AC = strtoint(&str, 16);
	if (*str)
		reg.XR = strtoint(&str, 16);
	if (*str)
		reg.YR = strtoint(&str, 16);
	if (*str)
		reg.SP = strtoint(&str, 16);
}

#if defined(APE_POSIX)
/*
 * TEDMON syntax is S "FILENAME" DEVICE   ADDRESS1 ADDRESS2
 *
 * APE65  syntax is S "FILENAME" ADDRESS1 ADDRESS2
 *
 * Returns
 * 	false		Okay
 * 	true		Syntax error
 */
static uint8_t
save(char *str)
{
	FILE *fp;
	char fname[SLEN];

	uint16_t f;
	uint16_t l;

	if (getfilename(&str, SLEN, fname) == NULL)
		return (true);

	/* First address */
	if (*str)
		f = strtoint(&str, 16);
	else
		return (true);

	/* Last address */
	if (*str)
		l = strtoint(&str, 16);
	else
		return (true);
	
	if ((fp = fopen(fname, "wb")) == NULL) {
		printf("I/O ERROR\r\n");
		return (false);
	}

	mon.addr = f;
	printf("SAVING %s\r\n", fname);

	fputc(f & 0xFF, fp);
	fputc(f >> 8, fp);
	while (f != l && getrunstop() != 3)
		fputc(peek(f++), fp);
	fclose(fp);

	return (false);
}

/*
 * TEDMON syntax is L "FILENAME" DEVICE
 *
 * APE65  syntax is L "FILENAME" [ADDRESS]
 *
 * Returns
 * 	false		Okay
 * 	true		Syntax error
 */
static uint8_t
load(char *str)
{
	FILE *fp;
	int c;
	char fname[SLEN];
	struct stat st;

	uint16_t f;

	if (getfilename(&str, SLEN, fname) == NULL)
		return (true);

	printf("SEARCHING FOR %s\r\n", fname);

	if (stat(fname, &st) != 0) {
		printf("I/O ERROR\r\n");
		return (false);
	}

	if (!S_ISREG(st.st_mode)) {
		printf("I/O ERROR\r\n");
		return (false);
	}

	if ((fp = fopen(fname, "rb")) == NULL) {
		printf("I/O ERROR\r\n");
		return (false);
	}

	/* Raw address */
	if (*str) {
		f = strtoint(&str, 16);
	} else {
		f = fgetc(fp);
		f |= (fgetc(fp) << 8);
	}

	printf("LOADING AT %04X\r\n", f);

	while ((c = fgetc(fp)) != EOF && getrunstop() != 3) {
		poke(f++, c);
	}
	fclose(fp);

	return (false);
}

/*
 * TEDMON syntax is V "FILENAME" DEVICE
 *
 * APE65  syntax is V "FILENAME" [ADDRESS]
 *
 * Returns
 * 	false		Okay
 * 	true		Syntax error
 */
static uint8_t
verify(char *str)
{
	FILE *fp;
	int c;
	char fname[SLEN];
	struct stat st;

	uint16_t f;
	uint16_t e = 0;

	if (getfilename(&str, SLEN, fname) == NULL)
		return (true);

	printf("SEARCHING FOR %s\r\n", fname);

	if (stat(fname, &st) != 0) {
		printf("I/O ERROR\r\n");
		return (false);
	}

	if (!S_ISREG(st.st_mode)) {
		printf("I/O ERROR\r\n");
		return (false);
	}

	if ((fp = fopen(fname, "rb")) == NULL) {
		printf("I/O ERROR\r\n");
		return (false);
	}

	/* Raw address */
	if (*str) {
		f = strtoint(&str, 16);
	} else {
		f = fgetc(fp);
		f |= (fgetc(fp) << 8);
	}

	printf("VERIFYING AT %04X ", f);

	while ((c = fgetc(fp)) != EOF && getrunstop() != 3) {
		if (peek(f++) != c)
			e++;
	}
	fclose(fp);

	if (e)
		printf("ERROR\r\n");
	else
		printf("OKAY\r\n");

	return (false);
}
#endif /* APE_POSIX */

#if 0
/*
 * Returns
 * 	false		Okay
 * 	true		Error
 */
static uint8_t
list_close(void)
{
	if (mon.list == NULL)
		return (true);

	fclose(mon.list);
	mon.list = NULL;

	return (false);
}

/*
 * Returns
 * 	false		Okay
 * 	true		Error
 */
static uint8_t
list_open(char *str)
{
	char fname[SLEN];

	list_close();

	if (getfilename(&str, SLEN, fname) == NULL)
		return (true);

	if ((mon.list = fopen(fname, "wb")) == NULL) {
		printf("I/O ERROR\r\n");
		return (true);
	}

	return (false);
}
#endif

/*
 * Returns
 * 	false		Okay
 * 	true		Syntax error
 */
static uint8_t
assemble(char *str)
{
	uint16_t addr;
	uint8_t bytes[3];
	uint8_t nbytes;
	char asmstr[SLEN];

	/* Address */
	if (*str)
		addr = strtoint(&str, 16);
	else
		return (true);

	strcpy(asmstr, str);

	while (assembler(addr, asmstr, bytes, &nbytes) == false) {

		/*
		 * Display opcode and data
		 */
		disassembler(addr, asmstr, bytes, &nbytes);
		printf("\033[1A\033[0KA %04X  %s\r\n", addr, asmstr);

		/*
		 * Store opcode and data
		 */
		switch (nbytes) {
		case 3: poke(addr + 2, bytes[2]);
		case 2: poke(addr + 1, bytes[1]);
		case 1: poke(addr + 0, bytes[0]);
		default:break;
		}

		/*
		 * Next line
		 */
		addr += nbytes;
		printf("A %04X ", addr);
		fflush(stdout);

		getstring(asmstr, SLEN);

		/* Finished? */
		if (*asmstr == '\0')
			return (false);
	}
	
	return (true);
}

static void
disassemble(char *str)
{
	uint16_t f;
	uint16_t l;
	uint8_t found;

	uint8_t bytes[3];
	uint8_t nbytes;

	/* First address */
	if (*str)
		f = strtoint(&str, 16);
	else
		f = mon.addr;

	/* Last address */
	if (*str)
		l = strtoint(&str, 16);
	else
		l = f + 0x14;

	found = false;
	while (found == false && getrunstop() != 3) {
		bytes[0] = peek(f);
		bytes[1] = peek(f + 1);
		bytes[2] = peek(f + 2);
		disassembler(f, str, bytes, &nbytes);
		printf(". %04X  %s\r\n", f, str);
		while (nbytes--)
			if (f++ == l)
				found = true;
	}

	mon.addr = f;
}

static void
execute(char *str)
{
	if (*str)
		reg.PC = strtoint(&str, 16);

	if (mon.trace)
		trace();
	else while (mpu() == 0 && getrunstop() != 3)
		;

	if (runstop == 3)
	        printf("\r\n");
        registers(BANNER | REGISTERS | NEWLINE);
        mon.addr = reg.PC;
}

static void
basic(void)
{
	reg.PC = 0xFF80;

	while (mpu() == 0 && runstop != 19)
			;

	printf("\r\n");
	registers(BANNER | REGISTERS | NEWLINE);
	mon.addr = reg.PC;
}

static uint8_t
command(char *str)
{
	uint8_t quit = false, err = false;
	char cmd;

	runstop = false;

	cmd = *str++;
	skipspace(&str);

	if (cmd >= 'a' && cmd <= 'z')
		cmd -= ('a' - 'A');

	switch (cmd) {
	default:
		err = true;
		break;
	case '?':
		/* Help */
		printf( "A assemble     D disassemble  \r\n"
			"C compare      T transfer     \r\n"
			"F fill         H hunt         \r\n"
			"G go           R register dump\r\n"
			"M memory dump  > modify memory\r\n"
			"; modify reg   X exit         \r\n"
#if defined(APE_POSIX)
			"L load file    S save file    \r\n"
			"V verify file  ! shell command\r\n"
#endif
#if 0
			"P list file on Q list file off\r\n"
			"N nmi          I irq          \r\n"
#endif
			"+ trace on     - trace off    \r\n"

			"$ hex convert  # dec convert  \r\n"
			"%% bin convert  @ oct convert  \r\n"

			"B start BASIC ^S stop BASIC   \r\n"
			);
		break;
	case 'R':
		/* Display registers */
		registers(BANNER | REGISTERS | NEWLINE);
		break;
#if 0
	case 'I':
		/* IRQ */
		irq();
		break;
	case 'N':
		/* NMI */
		nmi();
		break;
#endif
	case '#':
		/* Decimal base conversion */
		convert(strtoint(&str, 10));
		break;
	case '$':
		/* Hexadecimal base conversion */
		convert(strtoint(&str, 16));
		break;
	case '@':
		/* Octal base conversion */
		convert(strtoint(&str, 8));
		break;
	case '%':
		/* Binary base conversion */
		convert(strtoint(&str, 2));
		break;
	case '+':
		/* Trace on */
		printf("TRACE ON\r\n");
		mon.trace = true;
		break;
	case '-':
		/* Trace off */
		printf("TRACE OFF\r\n");
		mon.trace = false;
		break;
	case 'M':
		/* Memory display */
		memory(str);
		break;
	case 'F':
		/* Fill memory */
		err = fill(str);
		break;
	case '>':
		/* Modify memory */
		modify(str);
		break;
	case 'T':
		/* Transfer memory */
		err = transfer(str);
		break;
	case 'C':
		/* Compare memory */
		err =compare(str);
		break;
	case 'H':
		/* Hunt memory */
		err = hunt(str);
		break;
	case ';':
		/* Modify register(s) */
		setreg(str);
		break;
#if defined(APE_POSIX)
	case 'S':
		/* Save to disk */
		err = save(str);
		break;
	case 'L':
		/* Load from disk */
		err = load(str);
		break;
	case 'V':
		/* Verify on disk */
		err = verify(str);
		break;
	case '!':
		{
		/* Shell command */
		FILE *pf;
		char pstr[SLEN];
		int i;

		snprintf(pstr, SLEN, "%s 2>&1", str);
		pf = popen(pstr, "r");
		if (pf) {
			while (fgets(pstr, SLEN, pf) != NULL) {
				rtrim(pstr, SLEN);
				for (i = 0; pstr[i]; ++i)
					putcharacter(pstr[i]);
				putcharacter('\r');
				putcharacter('\n');
			}
			pclose(pf);
		} else {
			printf("ERROR\r\n");
		}
		}
		break;
#endif /* APE_POSIX */
#if 0
	case 'P':
		/* List file enable */
		err = list_open(str);
		break;
	case 'Q':
		/* List file disable */
		err = list_close();
		break;
#endif
	case 'X':
		/* Exit */
		quit = true;
		break;
	case '.':
	case 'A':
		/* Assemble */
		err = assemble(str);
		break;
	case 'D':
		/* Disassemble */
		disassemble(str);
		break;
	case 'G':
		/* Execute */
		execute(str);
		break;
	case 'B':
		/* BASIC */
		basic();
		break;
	}
	if (err)
		printf("\033[1A\033[0K?\r\n");

	return (quit);
}

void
monitor(void)
{
	char str[SLEN];

	mon.trace = false;
	mon.addr = 0;

	printf( "\r\n"
		"Welcome to the APE65 MCS6502 simulator\r\n"
		"--------------------------------------\r\n");

	printf("Written by Darron M. Broad.\r\n"
		"August 1989, February & March 2015.\r\n\r\n"
		);

	printf("Includes EhBASIC by Lee Davison.\r\n\r\n");

	printf("Enter ? for HELP.\r\n");

	printf("\r\n");

	reg.SR = UNKNOWN | BREAK;
	reg.SP = 0xFF;
	registers(BANNER | REGISTERS | NEWLINE);

	while (true) {
		getstring(str, SLEN);

		if (*str) {
			if (command(str))
				break;
		}
	}
}
