/*
 * Copyright (C) 2015 Darron Broad
 * All rights reserved.
 * 
 * This file is part of Ape65.
 * 
 * Ape65 is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation. 
 * 
 * Ape65 is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details. 
 * 
 * You should have received a copy of the GNU General Public License along
 * with Ape65. If not, see http://www.gnu.org/licenses/
 */

/*
 * 6502 simulator
 */

/*
 * Pinout
 * ------
 * !MCLR                                 1======28 AVDD 3V3
 * VREF+/CVREF+/AN0/C3INC/RPA0/CTED1/RA0 2      27 AVSS GND
 * VREF-/CVREF-/AN1/RPA1/CTED2/RA1       3      26 AN9/C3INA/RPB15/SCK2/CTED6/PMCS1/RB15
 * PGD1/AN2/C1IND/C2INB/C3IND/RPB0/RB0   4      25 CVREFOUT/AN10/C3INB/RPB14/SCK1/CTED5/PMWR/RB14
 * PGC1/AN3/C1INC/C2INA/RPB1/CTED12/RB1  5      24 AN11/RPB13/CTPLS/PMRD/RB13
 * AN4/C1INB/C2IND/RPB2/SDA2/CTED13/RB2  6      23 AN12/PMD0/RB12
 * AN5/C1INA/C2INC/RTCC/RPB3/SCL2/RB3    7      22 PGC2/TMS/RPB11/PMD1/RB11
 * VSS GND                               8      21 PGD2/RPB10/CTED11/PMD2/RB10
 * OSC1/CLKI/RPA2/RA2                    9      20 VCAP
 * OSC2/CLKO/RPA3/PMA0/RA3               10     19 VSS GND
 * SOSCI/RPB4/RB4                        11     18 TDO/RPB9/SDA1/CTED4/PMD3/RB9
 * SOSCO/RPA4/T1CK/CTED9/PMA1/RA4        12     17 TCK/RPB8/SCL1/CTED10/PMD4/RB8
 * VDD 3V3                               13     16 TDI/RPB7/CTED3/PMD5/INT0/RB7
 * PGD3/RPB5/PMD7/RB5                    14=====15 PGC3/RPB6/PMD6/RB6
 *
 * POWER
 * -----
 *    3V3--------VDD                    13, 28
 *
 *     +---------VCAP                   20
 *     |
 * 10u = E.S.R. < 5 OHMS
 *     |
 *     +---------VSS                    8, 19, 27
 *     |
 *    ///
 *
 * ICSP
 * ----
 * !MCLR        PROGRAMMING ENABLE      1
 * PGC1         PRIMARY CLOCK           5
 * PGD1         PRIMARY DATA            4
 * PGC2         SECONDARY CLOCK         22
 * PGD2         SECONDARY DATA          21
 * PGC3         TERTIARY CLOCK          15
 * PGD3         TERTIARY DATA           14
 */

#include "ape65.h"

#if defined(APE_MX150) || defined(APE_MX250)
#pragma config FPLLIDIV = DIV_2 /*  8 /  2 */
#pragma config FPLLMUL = MUL_20 /*  4 * 20 */
#pragma config FPLLODIV = DIV_2 /* 80 /  2 */
#pragma config FWDTEN = OFF
#pragma config FNOSC = FRCPLL
#pragma config FSOSCEN = OFF	/* RB4 I/O */
#pragma config JTAGEN = OFF
#endif

#if defined(APE_MX170) || defined(APE_MX270)
#pragma config FPLLIDIV = DIV_2 /*  8 /  2 */
#pragma config FPLLMUL = MUL_24 /*  4 * 24 */
#pragma config FPLLODIV = DIV_2 /* 96 /  2 */
#pragma config FWDTEN = OFF
#pragma config FNOSC = FRCPLL
#pragma config FSOSCEN = OFF	/* RB4 I/O */
#pragma config JTAGEN = OFF
#endif

uint8_t *memory = NULL;

extern const uint8_t BASIC[BASIC_SIZE];

extern const uint8_t MONITOR[MONITOR_SIZE];

uint8_t
peek(uint16_t address)
{
	if (address < RAM_SIZE) {
		return memory[address];
	}
	if (address >= BASIC_ADDR && address < (BASIC_ADDR + BASIC_SIZE)) {
		return BASIC[address - BASIC_ADDR];
        }
	if (address >= MONITOR_ADDR) {
		return MONITOR[address - MONITOR_ADDR];
        }
	if (address == 0xF004) {
		return getcharacter();
	}
	if (address == 0xFD10) {
		return PORTB;
	}
	return 0xFF;
}

void
poke(uint16_t address, uint8_t value)
{
	if (address < RAM_SIZE) {
		memory[address] = value;
		return;
	}
	if (address == 0xF001) {
		putcharacter(value);
		return;
	}
	if (address == 0xFD10) {
		TRISB = (TRISB & 0xFF00) | (value);
		return;
	}
}

static inline void
init_pps(void)
{
	/* TX2 = RB9(18) RX2 = RB8(17) */
	PPSUnLock;
	PPSOutput(4, RPB9, U2TX);
	PPSInput(2, U2RX, RPB8);
	PPSLock;
}

static inline void
init_io(void)
{
	/* ADC Digital Mode */
	ANSELA = 0;
	ANSELB = 0;

	/*
	 * PORT A
	 *
	 * RA0 connects to a LED and RA1 a piezo sounder.
	 */

	/* RESET */
	LATA = 0x01;
	/* RA0/1 O/P */
	TRISAbits.TRISA0 = 0;
	TRISAbits.TRISA1 = 0;

	/*
	 * PORT B
	 *
	 * This port emulates the MOS 6529B as best it can.
	 *
	 * The method used is TRIS open-drain emulation.
	 * As per the original chip, a bit in TRIS is
	 * set to 1 if we want to use it as an input and
	 * set to 0 if we want to set the output to 0.
	 * If we want to set an output to 1, then this
	 * depends on an external pull-up resistor to
	 * provide the power. This method only works
	 * if the output latch is always 0.
	 */

	/* RESET */
	LATB = 0x00;
	/* RB9 O/P */
	TRISB = 0xFDFF;
	/* WEAK PULL-UP DISABLED */
	CNPUB = 0x0000;
	/* WEAK PULL-DOWN DISABLED */
	CNPDB = 0x0000;
	/* OPEN-DRAIN DISABLED */
	ODCB  = 0x0000;
}

/*
 * Blink LED n times whilst activating piezo buzzer
 */
#define STEP (5)
static inline void
blink(uint32_t n)
{
	uint16_t i;

	n <<= 1;
	while (n--) {
		/* BUZZ RA1 */

		for (i = 0; i < 250; i += STEP) {
			delay_ms(STEP);

			/* RA1 = ! RA1 */
			LATAbits.LATA1 = ~LATAbits.LATA1;
		}

		/* TOGGLE RA0 AT 2 Hz */

		/* RA0 = ! RA0 */
		LATAbits.LATA0 = ~LATAbits.LATA0;
	}
}
#undef STEP

int
main(int argc, char **argv)
{
	SYSTEMConfigPerformance(FCY);

	init_pps();
	init_io();
	init_uart2();

	if ((memory = malloc(RAM_SIZE)) == NULL) {
		printf("\r\nsystem initialisation error\r\n");
		while (1)
			;
	}
	for (uint32_t i = 0; i < RAM_SIZE;) {
		poke(i++, 0x00);
		poke(i++, 0xFF);
	}

	blink(2);

	/* Simulate PLUS/4 TEDMON */
	while (1)
		monitor();
}
