/*
 * Copyright (C) 2015 Darron Broad
 * All rights reserved.
 * 
 * This file is part of Ape65.
 * 
 * Ape65 is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation. 
 * 
 * Ape65 is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details. 
 * 
 * You should have received a copy of the GNU General Public License along
 * with Ape65. If not, see http://www.gnu.org/licenses/
 */

/*
 * 6502 simulator
 */

#include "ape65.h"

uint8_t *memory = NULL;

extern const uint8_t BASIC[BASIC_SIZE];

extern const uint8_t MONITOR[MONITOR_SIZE];

struct termios old_options, new_options;

char runstop = '\0';

uint8_t
peek(uint16_t address)
{
	if (address < BASIC_ADDR) {
		return memory[address];
	}
	if (address < (BASIC_ADDR + BASIC_SIZE)) {
		return BASIC[address - BASIC_ADDR];
	}
	if (address >= MONITOR_ADDR) {
		return MONITOR[address - MONITOR_ADDR];
	}
	if (address == 0xF004) {
		return getcharacter();
	}
	return 0xFF;
}

void
poke(uint16_t address, uint8_t value)
{
	static uint8_t x = 0, y = 0;

	if (address < BASIC_ADDR) {
		memory[address] = value;
		return;
	}
	if (address == 0xF001) {
		putcharacter(value);
		return;
	}
	if (address == 0xF005) {
		char str[SLEN];
		snprintf(str, SLEN, "\033[%d;%dH", x = value, y);
		write(STDOUT_FILENO, str, strlen(str));
		return;
	}
	if (address == 0xF006) {
		char str[SLEN];
		snprintf(str, SLEN, "\033[%d;%dH", x, y = value);
		write(STDOUT_FILENO, str, strlen(str));
		return;
	}
}

static inline void
init_stdio(void)
{
	/* Get options */
	tcgetattr(STDIN_FILENO, &old_options);

	new_options = old_options;

	/*
	 * Raw mode
	 *
	 *  Linux TERMIOS(3)
	 */

	/* Input options */
	new_options.c_iflag &= ~(IGNBRK | BRKINT | PARMRK | ISTRIP | INLCR | IGNCR | ICRNL | IXON);

	/* Output options */
	new_options.c_oflag &= ~(OPOST);

	/* Control options */
	new_options.c_cflag &= ~(CSIZE | PARENB);
	new_options.c_cflag |= (CS8);

	/* Local options */
	new_options.c_lflag &= ~(ECHO | ECHONL | ICANON | ISIG | IEXTEN);

	/* Control characters */
	new_options.c_cc[VMIN] = 0;
	new_options.c_cc[VTIME] = 0;

	/* Baud rate */
	cfsetispeed(&new_options, B9600);
	cfsetospeed(&new_options, B9600);

	/* Set options */
	tcsetattr(STDIN_FILENO, TCSANOW, &new_options);
}

char
getcharacter(void)
{
	struct timeval tv = {0, 10};
	fd_set readfds;
	char c = 0;

	FD_ZERO(&readfds);
	FD_SET(STDIN_FILENO, &readfds);

	if (select(STDIN_FILENO + 1, &readfds, NULL, NULL, &tv) == 1) {
		if (read(STDIN_FILENO, &c, sizeof(char)) < 0)
			return (0);
		if (c >= 'A' && c <= 'Z')
			c += 'a' - 'A';
		else if (c >= 'a' && c <= 'z')
			c -= 'a' - 'A';
		else if (c == 127)
			c = 8;
		runstop = c;
	}

	return (c);
}

void
putcharacter(char c)
{
	if (write(STDOUT_FILENO, &c, sizeof(char)) < 0) {
		/* Don't care */
	}
}

int
main(int argc, char **argv)
{
	init_stdio();

	if ((memory = malloc(0xC000)) == NULL) {
		printf("\r\nsystem initialisation error\r\n");
		exit(1);
	}
	for (uint32_t i = 0; i < 0xC000;) {
		poke(i++, 0x00); 
		poke(i++, 0xFF);
	}   

	/* Simulate PLUS/4 TEDMON */
	monitor();

	free(memory);

	tcsetattr(STDIN_FILENO, TCSANOW, &old_options);

	exit(EX_OK);
}
